import desktop

import gtk


#
# Class for transparent windows.
#
class GlassWindow(gtk.Window):

    # transparency modes
    TRANSPARENCY_NONE = 0
    TRANSPARENCY_SCREENSHOT = 1
    TRANSPARENCY_PSEUDO = 2

    # window types
    TYPE_NORMAL = 0
    TYPE_KEEP_BELOW = 1
    TYPE_KEEP_ABOVE = 2
    TYPE_STICKY = 4
    TYPE_MANAGED = 8
    

    #
    # Constructor.
    #
    def __init__(self, wintype = gtk.WINDOW_TOPLEVEL):

        # the transparency mode to use
        self.__transparency = self.TRANSPARENCY_PSEUDO

        # the desktop background
        self.__bg_root = None

        # remember the current window position to detect movements
        self.__position = (0, 0)
        self.__size = (0, 0)

        # whether the window is below all others
        self.__is_lower = 0
        self.__lowering_lock = 0

        self.__is_above = 0

        gtk.Window.__init__(self, wintype)
        self.connect("configure-event", self.__on_configure)
        desktop.watch_bg(self.__bg_observer)

        self.__bg_image = gtk.gdk.Pixbuf(0, 1, 8, 10, 10)
        self.__bg_image.fill(0xffffff)

        
        self.connect("focus-in-event", self.__lower)
        self.connect("button-press-event", self.__lower)
        #self.connect("expose-event", self.__lower) # caused trouble with XFCE4



    def __lower(self, src, event):

        if (self.__lowering_lock): return
        
        if (self.__is_lower and self.window):
            self.window.lower()


        # lock lowering for some time to avoid CPU intensive flickering
        self.__lowering_lock = 1
        def f(self): self.__lowering_lock = 0
        gtk.timeout_add(25, f, self)



    def __raise(self, src, event):

        x, y = self.get_pointer()
        w, h = self.size_request()
        
        if (self.window):
            self.window.raise_()

        if (self.__is_above):
            gtk.timeout_add(100, self.__raise, None, None)




    #
    # Observer method for the background.
    # Connect this method to the BG watcher.
    #
    def __bg_observer(self, src, cmd):

        self.update_bg()




    #
    # Sets the type of this window. Multiple types can be ORed together.
    #
    def set_window_type(self, wtype):

        self.__is_lower = 0
        self.__is_above = 0


        if (wtype & self.TYPE_KEEP_BELOW):
            self.__is_lower = 1
            self.__lower(None, None)
        elif (wtype & self.TYPE_KEEP_ABOVE):
            self.__is_above = 1
            self.__raise(None, None)
        #else:
        #    pass


        if (wtype & self.TYPE_STICKY):
            self.stick()
        else:
            self.unstick()

        if (wtype & self.TYPE_MANAGED):
            self.set_decorated(gtk.TRUE)
        else:
            self.set_decorated(gtk.FALSE)



    #
    # Sets the transparency mode of this window.
    #
    def set_transparency_mode(self, mode):

        self.__transparency = mode

        if (mode == self.TRANSPARENCY_NONE):
            self.__bg_root = gtk.gdk.Pixbuf(0, 1, 8, width, height)
            self.__bg_root.fill(0xffffff)



    #
    # Displays the background image.
    #
    def __display_bg(self):

        if (not self.__bg_root): return

        width = self.__bg_root.get_width()
        height = self.__bg_root.get_height()

        # we cannot use Pixbuf.copy() here because of a memory leak
        # bug in pygtk versions < 1.99.14
        image = gtk.gdk.Pixbuf(0, 1, 8, width, height)
        self.__bg_root.copy_area(0, 0, width, height, image, 0, 0)


        # set as window bg
        pix, mask = image.render_pixmap_and_mask()

        style = gtk.Style()
        style.bg_pixmap[gtk.STATE_NORMAL] = pix
        self.set_style(style)

        # tell gtk to redraw this window ASAP
        self.queue_draw()




    #
    # Updates the background for transparency.
    #
    def update_bg(self, noscreenshot = 0):

        if (not self.window): return
        
        x, y = self.window.get_origin()
        width, height = self.window.get_size()
        if (self.__transparency != self.TRANSPARENCY_NONE
            and width > 1 and height > 1):
            self.__capture_bg(x, y, width, height, noscreenshot)
            self.__display_bg()



    #
    # Captures the background to create transparency.
    #
    def __capture_bg(self, x, y, width, height, noscreenshot = 0):

        # try capturing the background
        if (self.__transparency == self.TRANSPARENCY_PSEUDO):
            try:
                self.__bg_root = desktop.get_wallpaper(x, y, width, height)
                return
            except:
                try:
                    self.__bg_root = desktop.get_wallcolor(width, height)
                    return
                except:
                    self.__bg_root = gtk.gdk.Pixbuf(0, 1, 8, width, height)
                    self.__bg_root.fill(0xddddee)
                    return
                #end try
            #end try
        #end if

        # use screenshot transparency if capturing has failed or transparency
        # mode is set to screenshot transparency
        if (noscreenshot): return  # must not make screenshot now

        self.hide()
        self.move(-1000, 0)
        self.queue_draw()

        # wait long enough for the window to disappear; does this work always?
        while (gtk.gdk.events_pending()): gtk.mainiteration()
        gtk.mainiteration()
        
        self.__bg_root = desktop.get_wallpaper_fallback(x, y, width, height)
        self.show()
        self.move(x, y)


    #
    # Reacts on moving the window.
    #
    def __on_configure(self, src, event):

        pos = self.window.get_origin()
        size = self.window.get_size()
        if (pos != self.__position or size != self.__size):
            self.__position = pos
            self.__size = size

            # don't make screenshots while moving the window
            self.update_bg(noscreenshot = 1)
